"""Defining Sensor ABC and its derived classes."""

from __future__ import annotations
import random
import logging
from pydantic import BaseModel
import smbus2  # or just smbus
from gpiozero import DigitalInputDevice
from bairy.device.utils import configure_logging
from bairy.device.configs import LOG_PATH

configure_logging(LOG_PATH)


class Sensor:
  """An abstract base class for sensors."""

  def __init__(self, configs: BaseModel, prev_reading: int | None = None):
    for k, v in configs.dict().items():
      setattr(self, k, v)
    self.prev_reading = prev_reading

    if self.sensor_type == 'digital':
      self.device = DigitalInputDevice(self.bcm_pin)

  def read(self) -> dict[str, int | None]:
    """Read sensor measurements and return dictionary of values."""
    read_dict = {'air': self.read_air,
                 'random': self.read_random,
                 'digital': self.read_digital}
    return read_dict[self.sensor_type]()

  def read_air(self):
    """Read I2C data from air sensor."""

    keys_dict = {
        'pm_1.0': (4, 5),
        'pm_2.5': (6, 7),
        'pm_10': (8, 9),
        'n_beyond_0.3': (16, 17),
        'n_beyond_0.5': (18, 19),
        'n_beyond_1.0': (20, 21),
        'n_beyond_2.5': (22, 23),
        'n_beyond_5.0': (24, 25),
        'n_beyond_10': (26, 27)}
    data_dict: dict[str, int | None] = {k: None for k in keys_dict}

    with smbus2.SMBus(1) as bus:
      try:
        data: list[int] = bus.read_i2c_block_data(self.i2c_address, 0, 32)
      except OSError:  # couldn't read data -- sensor disconnected?
        logging.warning('Unable to connect to air sensor over I2C')
        return data_dict

      # start characters, check sum, error byte
      try:
        assert data[0] == 0x42
        assert data[1] == 0x4d
        assert data[29] == 0
      except AssertionError:
        logging.warning('Internal error on air sensor.')
      if sum(data[:30]) != (data[30] << 8) + data[31]:
        logging.warning('Bad checksum from air sensor.')

      for k, (byte1, byte2) in keys_dict.items():
        data_dict[k] = (data[byte1] << 8) + data[byte2]
    return data_dict

  def read_digital(self):
    """Read value associated to generic digital sensor."""
    try:
      v: int | None = self.device.value
    except RuntimeError:
      logging.warning('RuntimeError on sensor with header: %s', self.header)
      v = None

      # try closing then reopening
      self.device.close()
      self.device = DigitalInputDevice(self.bcm_pin)

    return {self.header: v}

  def read_random(self):
    """Create random data for testing device."""
    if self.prev_reading is None:
      r = random.randint(0, 255)
    else:
      r = self.prev_reading + random.randint(-3, 3)
      r = max(min(r, 255), 0)  # clipping
    self.prev_reading = r
    return {self.header: r}



"""Gather configuration and data from devices."""

from __future__ import annotations
import logging
import os
import asyncio
import nest_asyncio
import aiohttp
from bairy.hub import configs
from bairy.device import utils, configs as device_configs
from bairy.device.app import status_json

nest_asyncio.apply()
utils.configure_logging(configs.LOG_PATH)


async def get_status(ip_address: str):
  """Get status of device associated to ip_address."""
  if ip_address == 'self':
    return status_json()
  async with aiohttp.ClientSession() as session:
    url = 'http://' + ip_address + ':8000/' + 'status.json'
    async with session.get(url) as r:
      return await r.json()


def get_all_statuses():
  """Get status of every device known to hub."""
  ip_addresses = configs.load_ips()
  loop = asyncio.get_event_loop()
  tasks = [get_status(ip_address) for ip_address in ip_addresses]
  gathered = asyncio.gather(*tasks)
  statuses = loop.run_until_complete(gathered)
  return statuses


def validate_names():
  """Check device names to guarantee no duplicates."""
  statuses = get_all_statuses()
  names = [s['device_configs']['name'] for s in statuses]
  if len(set(names)) < len(list(names)):
    raise ValueError('Discovered repeated name within devices!')
  logging.info('Validated device names %s', names)


async def get_data(ip_address: str):
  """Request /data endpoint from device and save response to file."""
  try:
    status = await get_status(ip_address)
    name: str = status['device_configs']['name']
    data_path = os.path.join(configs.HUB_DATA_DIR, name + '.csv')

    if ip_address == 'self':  # making a symlink to data
      os.symlink(device_configs.DATA_PATH, data_path)

    url = 'http://' + ip_address + ':8000/data'

    # make a backup copy of existing data
    backup_path = os.path.join(configs.BACKUP_DIR, name + '.csv')
    if os.path.exists(data_path):
      os.rename(data_path, backup_path)

    logging.info('Requesting data from %s', ip_address)
    await stream_request(url, data_path)

    # comparing the number of rows of the new data with the old data
    size = utils.count_rows(data_path)
    if os.path.exists(backup_path):
      if size < utils.count_rows(backup_path):
        raise ValueError('New data is missing some of previous data.')
      os.remove(backup_path)
    logging.info('Successfully saved data from %s', ip_address)

  except aiohttp.ClientConnectionError as e:
    logging.error('Failed to connect to %s', ip_address)
    logging.error(e)


async def stream_request(url: str, save_path: str):
  """Make stream request and save data with aiohttp."""
  async with aiohttp.ClientSession() as session:
    async with session.get(url) as r:
      with open(save_path, 'wb') as f:
        while True:
          chunk = await r.content.read(1024)
          if not chunk:
            break
          f.write(chunk)


async def request_data_indefinitely(ip_address: str):
  """Request data from each device indefinitely."""
  assert ip_address != 'self'
  while True:
    await get_data(ip_address)
    await asyncio.sleep(configs.RECACHE_INTERVAL)


def run_requests():
  """Run requests indefinitely."""
  loop = asyncio.get_event_loop()
  for ip_address in configs.load_ips():
    if ip_address != 'self':
      loop.create_task(request_data_indefinitely(ip_address))
  loop.run_forever()



"""Control IoT device by reading sensor values and writing data to disk."""

from __future__ import annotations
from bairy.device.validate import DeviceConfigs
import os
import asyncio
from datetime import datetime
from bairy.device.configs import DATA_PATH, DATE_FORMAT, load_device
from bairy.device.sensor import Sensor


def read_sensors(sensors: list[Sensor]):
  """Read sensor values."""
  data: dict[str, int | None] = {}
  for s in sensors:
    reading = s.read()
    for k in reading:
      if k in data:
        raise KeyError('Duplicate key found!')
    data.update(reading)
  return data


def write_data(data: dict[str, int | None]):
  """Create a data file if none exists and append data to end."""
  time = datetime.now().strftime(DATE_FORMAT)
  values_as_str = [str(v) if v is not None else '' for v in data.values()]
  row = time + ',' + ','.join(values_as_str) + '\n'
  with open(DATA_PATH, 'a') as f:
    f.write(row)


def create_data_file(sensors: list[Sensor]):
  """Create data file if none exists and write column headers, or return most recent data."""

  # taking an initial reading to get header values
  data = read_sensors(sensors)
  headers = data.keys()

  if not os.path.exists(DATA_PATH):
    headers = 'time,' + ','.join(headers) + '\n'
    with open(DATA_PATH, 'w') as f:
      f.write(headers)


def initialize_device(device: DeviceConfigs | None = None):
  """Helper function for run device."""
  if device is None:
    device = load_device()
  sensors = [Sensor(s) for s in device.sensors]
  create_data_file(sensors)
  return device, sensors


def run_device():
  """Run device indefinitely."""
  device, sensors = initialize_device()

  async def run():
    while True:
      data = read_sensors(sensors)
      write_data(data)
      await asyncio.sleep(device.update_interval)

  loop = asyncio.get_event_loop()
  loop.create_task(run())
  loop.run_forever()



"""Create dash plot and dash table as endpoints."""

from __future__ import annotations
import os
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
import plotly.io as pio
from dash import Dash
from dash_table import DataTable
import dash_core_components as dcc
import dash_html_components as html
from bairy.device import configs


pio.templates.default = 'plotly_white'


def determine_plot_configs():
  """Determine plotly axes and variables configurations."""

  d = configs.load_device()
  # the order of the dictionaries below matter for plot tracing
  # order is least important to most important
  sensor_headers: dict[str, list[str]] = {'random': [],
                                          'digital': [],
                                          'air': []}
  sensor_units = {'random': 'random',
                  'digital': 'intensity',
                  'air': 'micrograms / cubic meter'}

  for s in d.sensors:
    if s.sensor_type == 'air':
      sensor_headers['air'] += ['pm_1.0', 'pm_2.5', 'pm_10']
    else:
      sensor_headers[s.sensor_type].append(s.header)

  # cleanup dictionaries
  empties = [k for k in sensor_headers.keys() if sensor_headers[k] == []]
  for k in empties:
    del sensor_headers[k]
    del sensor_units[k]

  # keeping at most two types of headers
  if len(sensor_headers) == 3:
    del sensor_headers['random']
    del sensor_units['random']

  return sensor_headers, sensor_units


def preprocess_df(as_str: bool = False, only_last_day: bool = False):
  """Preprocess pandas DataFrame."""
  df = pd.read_csv(configs.DATA_PATH)

  df = df.set_index('time')
  df.index = pd.to_datetime(df.index)

  sensor_headers, _ = determine_plot_configs()
  cols_to_keep = [col for key in sensor_headers for col in sensor_headers[key]]
  df = df[cols_to_keep]

  if only_last_day:
    start = pd.Timestamp.now() - pd.Timedelta('1 day')
    df = df[df.index > start]

  df = resample_df(df)

  if as_str:  # for plotly table
    df = df.iloc[::-1]
    df.index = df.index.astype(str)
    df = df.round(3)
  return df.reset_index()  # move time back as a column


def resample_df(df):
  """Smooth and condense df by resampling."""
  # 60 has many divisors
  rules = ['1T', '2T', '3T', '4T', '5T', '6T', '10T', '20T', '30T', '1H']

  if len(df) > 300:  # only resample if df large enough
    for r in rules:
      df = df.resample(r).mean()
      if len(df) < 5000:
        break

  # smoothing even more
  df = df.rolling(7, center=True).mean()
  return df


def create_fig(only_last_day: bool = False):
  """Create plotly figure using one or two y-axes."""

  fig = go.Figure()
  if not os.path.exists(configs.DATA_PATH) or not os.path.exists(configs.CONFIGS_PATH):
    return fig

  # see https://plotly.com/python/discrete-color/
  colors = iter(px.colors.qualitative.Bold)
  df = preprocess_df(only_last_day=only_last_day)
  sensor_headers, sensor_units = determine_plot_configs()

  for i, (key, cols) in enumerate(sensor_headers.items()):
    unit = sensor_units[key]
    if len(sensor_headers) == 2 and i == 0:
      side = 'right'
      opacity = 0.4
    else:
      side = 'left'
      opacity = 1.0

    yaxis = 'y' if i == 0 else 'y2'
    yaxis_layout = 'yaxis' if i == 0 else 'yaxis2'
    layout_params = {yaxis_layout: {'title': unit, 'side': side}}
    if i == 1:
      layout_params[yaxis_layout]['overlaying'] = 'y'

    for col in cols:
      fig.add_trace(go.Scatter(
          x=df['time'],
          y=df[col],
          name=col,
          opacity=opacity,
          line={'color': next(colors)},
          yaxis=yaxis))
    fig.update_layout(**layout_params)

  # threshold for pm2.5
  if 'air' in sensor_headers and df['pm_2.5'].max() > 40:
    yaxis = 'y2' if len(sensor_headers) == 2 else 'y'
    fig.add_trace(go.Scatter(
        x=df['time'],
        y=[35] * len(df),
        name='pm_2.5 safe threshold',
        yaxis=yaxis,
        line={'dash': 'dot', 'color': next(colors)}))

  fig.update_layout(height=800)
  return fig


def serve_plot():
  """Dynamically serve dash_plot.layout."""

  fig_day = create_fig(only_last_day=True)
  fig_day.layout.title = 'Data from last 24 hours'
  fig_day.layout.xaxis.rangeslider.visible = True

  fig_all = create_fig(only_last_day=False)
  fig_all.layout.title = 'All available data'
  fig_all.layout.xaxis.rangeslider.visible = True

  return html.Div(children=[
      html.Div(children=[
          html.H1(children='bairy'),
          html.Div(children='Display sensor data from Raspberry Pi.'),
          html.A(
              'about bairy',
              href='https://github.com/zebengberg/bairy',
              style={'margin': '20px'}),
          html.A(
              'table',
              href='/table',
              style={'margin': '20px'})]),
      dcc.Graph(id='graph_all', figure=fig_all),
      dcc.Graph(id='graph_day', figure=fig_day)])


css = 'https://codepen.io/chriddyp/pen/bWLwgP.css'
dash_plot = Dash(requests_pathname_prefix='/plot/', external_stylesheets=[css])
# See https://dash.plotly.com/live-updates
dash_plot.layout = serve_plot


def serve_table():
  """Dynamically serve updated dash_table.layout."""
  if not os.path.exists(configs.DATA_PATH) or not os.path.exists(configs.CONFIGS_PATH):
    return DataTable()

  df = preprocess_df(True, True)
  columns = [{'name': i, 'id': i} for i in df.columns]
  data = df.to_dict('records')
  return html.Div(children=[

      html.Div(children=[
          html.H1(children='bairy'),
          html.Div(children='Display sensor data from Raspberry Pi.'),
          html.A(
              'about bairy',
              href='https://github.com/zebengberg/bairy',
              style={'margin': '20px'}
          ),
          html.A(
              'plot',
              href='/plot',
              style={'margin': '20px'}
          )]),
      DataTable(
          id='table',
          columns=columns,
          data=data,
          style_cell=dict(textAlign='left'),
          style_header=dict(backgroundColor="paleturquoise"),
          style_data=dict(backgroundColor="lavender"))])


dash_table = Dash(requests_pathname_prefix='/table/',
                  external_stylesheets=[css])
# See https://dash.plotly.com/live-updates
dash_table.layout = serve_table



"""Process device requests with FastAPI."""


from __future__ import annotations
import os
import json
import subprocess
import sys
import logging
from typing import Any
from fastapi import FastAPI
from fastapi.responses import PlainTextResponse, StreamingResponse, RedirectResponse
from fastapi.middleware.wsgi import WSGIMiddleware
import uvicorn
from bairy.device import utils, configs, dash_app


def configure_app_logging(log_path: str):
  """Configure uvicorn default logging to save to file."""

  LC: dict[Any, Any] = uvicorn.config.LOGGING_CONFIG

  # changing default formatting
  LC['formatters']['access']['fmt'] = configs.LOG_FORMAT
  LC['formatters']['access']['datefmt'] = configs.DATE_FORMAT
  LC['formatters']['default']['fmt'] = configs.LOG_FORMAT
  LC['formatters']['default']['datefmt'] = configs.DATE_FORMAT

  # adding FileHandler to LOGGING_CONFIG
  LC['handlers']['default_file'] = {
      'formatter': 'default',
      'class': 'logging.FileHandler',
      'filename': log_path}
  LC['handlers']['access_file'] = {
      'formatter': 'access',
      'class': 'logging.FileHandler',
      'filename': log_path}

  # telling loggers to use the additional handler
  LC['loggers']['uvicorn']['handlers'].append('default_file')
  LC['loggers']['uvicorn.access']['handlers'].append('access_file')


utils.configure_logging(configs.LOG_PATH)
configure_app_logging(configs.LOG_PATH)
app = FastAPI()
app.mount('/plot', WSGIMiddleware(dash_app.dash_plot.server))
app.mount('/table', WSGIMiddleware(dash_app.dash_table.server))


@app.get('/')
async def root():
  """Redirect to plot."""
  return RedirectResponse(url='/plot')


@app.get('/data')
def data():
  """Return streaming response of CSV containing all data."""
  # cannot use with ... here
  f = open(configs.DATA_PATH, 'rb')
  return StreamingResponse(f, media_type='text/csv')


@app.get('/logs', response_class=PlainTextResponse)
async def logs():
  """Return app log as plain text."""
  with open(configs.LOG_PATH) as f:
    return f.read()


@app.get('/status', response_class=PlainTextResponse)
def status():
  """Return device status as plaintext json."""
  device_status = status_json()
  return json.dumps(device_status, indent=4)


@app.get('/status.json')
def status_json():
  """Return device status as raw json."""
  device_configs = configs.load_device().dict()
  size = utils.get_data_size()
  n_rows = utils.count_rows(utils.DATA_PATH)
  latest = utils.latest_data()
  device_status = {'device_configs': device_configs,
                   'data_details': {'file_size': size, 'n_rows': n_rows},
                   'latest_reading': latest}
  return device_status


@app.get('/experimental/{param}')
def experimental(param: str):
  """Run experimental param on device."""
  if param == 'reboot':
    logging.info('reboot attempt ...')
    os.system('sudo reboot')
    return 'fail'  # if here, the reboot failed

  if param == 'update':
    logging.info('bairy update requested!')
    command = [sys.executable, '-m', 'pip', 'install', '-U', 'bairy']
    try:
      subprocess.check_call(command)
      logging.info('bairy update success')
      return 'success'
    except subprocess.CalledProcessError as e:
      logging.info('bairy update failure')
      logging.info(e)
      return e

  else:
    return 'unknown param'


def run_app():
  """Run app with uvicorn."""
  uvicorn.run(app, host='0.0.0.0', port=8000)



"""Create bairy.service file for systemd scripts."""

import os
import subprocess
import getpass


def get_bairy_path():
  """Get path of bairy installed by pip."""
  p = subprocess.Popen(['which', 'bairy'],
                       stdout=subprocess.PIPE,
                       stderr=subprocess.PIPE)
  out, err = p.communicate()

  assert len(err) == 0
  return out.decode().rstrip()


def create_service():
  """Create bairy.service to be put into systemd."""

  path = get_bairy_path()
  user = getpass.getuser()
  content = f'''
  [Unit]
  Description=Run bairy at startup
  After=network-online.target time-sync.target
  User=pi

  [Service]
  Type=oneshot
  ExecStart={path}
  User={user}

  [Install]
  WantedBy=default.target
  '''

  with open('bairy.service', 'w') as f:
    f.write(content)
  move_service()


def move_service():
  """Move bairy.service to systemd directory."""
  path = '/etc/systemd/system'
  if os.path.exists(path):
    subprocess.run(['sudo', 'mv', 'bairy.service', path])
    subprocess.run(['sudo', 'systemctl', 'enable', 'bairy.service'])
    print('Successfully moved bairy.service to systemd directory.')
  else:
    print('No systemd directory found.')
    os.remove('bairy.service')



"""Read and define configurations and paths for device."""

from __future__ import annotations
import os
import json
from typing import Any
from bairy.device.validate import random_configs, DeviceConfigs


# creating data directory
MODULE_DIR = os.path.dirname(__file__)
PACKAGE_DIR = os.path.dirname(MODULE_DIR)
DATA_DIR = os.path.join(PACKAGE_DIR, 'data')
DEVICE_DATA_DIR = os.path.join(DATA_DIR, 'device')
CONFIGS_PATH = os.path.join(DEVICE_DATA_DIR, 'configs.json')
LOG_PATH = os.path.join(DEVICE_DATA_DIR, 'app.logs')
DATA_PATH = os.path.join(DEVICE_DATA_DIR, 'data.csv')
DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
LOG_FORMAT = '%(asctime)s - %(levelname)s - %(message)s'
if not os.path.exists(DATA_DIR):
  os.mkdir(DATA_DIR)
if not os.path.exists(DEVICE_DATA_DIR):
  os.mkdir(DEVICE_DATA_DIR)


def set_configs(path: str):
  """Verify and save device configs as json file within data directory."""
  with open(path) as f:
    configs: dict[str, Any] = json.load(f)
  d = DeviceConfigs(**configs)
  assert d == DeviceConfigs(**d.dict())
  with open(CONFIGS_PATH, 'w') as f:
    json.dump(configs, f, indent=4)
  print('Successfully validated and set configs.')


def set_random_configs():
  """Save random configs as json file within data directory."""
  d = random_configs()
  assert d == DeviceConfigs(**d.dict())
  with open(CONFIGS_PATH, 'w') as f:
    json.dump(d.dict(), f, indent=4)


def load_device():
  """Look for stored configs.json file."""
  if os.path.exists(CONFIGS_PATH):
    with open(CONFIGS_PATH) as f:
      configs: dict[str, Any] = json.load(f)
    d = DeviceConfigs(**configs)
    assert d == DeviceConfigs(**configs)
    return d
  else:
    raise FileNotFoundError('No configurations found! Run bairy --help')



"""Validate configs with pydantic. Sensor fields are chosen so that
d -> Device(**d.dict()) is an involution."""


# cannot use __future__ annotations with pydantic
from typing import List, Union, Dict
from pydantic import BaseModel, validator


class AirSensorConfigs(BaseModel):
  sensor_type: str = 'air'
  i2c_address: int

  @validator('sensor_type')
  def check_sensor_type(cls, value: str):
    assert value == 'air'
    return value


class DigitalSensorConfigs(BaseModel):
  sensor_type: str = 'digital'
  bcm_pin: int
  header: str

  @validator('sensor_type')
  def check_sensor_type(cls, value: str):
    assert value == 'digital'
    return value


class RandomSensorConfigs(BaseModel):
  sensor_type: str = 'random'
  header: str

  @validator('sensor_type')
  def check_sensor_type(cls, value: str):
    assert value == 'random'
    return value


class DeviceConfigs(BaseModel):
  """A Class holding configuration fields of the device."""
  name: str
  sensors: List[Union[AirSensorConfigs,
                      DigitalSensorConfigs,
                      RandomSensorConfigs]]
  update_interval: int


def random_configs():
  """Return a device with a random sensor."""
  s1 = RandomSensorConfigs(sensor_type='random', header='random1')
  s2 = RandomSensorConfigs(sensor_type='random', header='random2')
  s3 = RandomSensorConfigs(sensor_type='random', header='random3')
  d = DeviceConfigs(
      name='random sensors',
      sensors=[s1, s2, s3],
      update_interval=1)
  assert d == DeviceConfigs(**d.dict())
  return d


def example_configs():
  """Return an example of valid configs as json."""
  s1 = AirSensorConfigs(sensor_type='air', i2c_address=0x12)
  s2 = DigitalSensorConfigs(sensor_type='digital',
                            bcm_pin=17, header='ir_state')
  s3 = DigitalSensorConfigs(sensor_type='digital',
                            bcm_pin=27, header='sound_state')
  d = DeviceConfigs(
      name='example sensors',
      sensors=[s1, s2, s3],
      update_interval=1)
  assert d == DeviceConfigs(**d.dict())
  return d



"""Test __main__ argparse."""

import os
from bairy.__main__ import parse_args
from bairy import create_service


def test_parser():
  """Test argparse command line arguments."""
  args = parse_args(['--print-configs'])
  assert args.print_configs

  args = parse_args(['--print-data-path'])
  assert args.print_data_path

  args = parse_args(['--create-service'])
  assert args.create_service

  args = parse_args(['--set-random-configs'])
  assert args.set_random_configs

  args = parse_args(['--set-configs', 'fake_path'])
  assert args.configs_path == ['fake_path']

  args = parse_args(['--remove', 'logs'])
  assert args.remove == ['logs']


def test_service():
  create_service.create_service()



"""Dash app to plot device data."""

import os
import glob
import pandas as pd
import plotly.express as px
from dash import Dash
import dash_core_components as dcc
import dash_html_components as html
from bairy.hub.configs import HUB_DATA_DIR, IP_PATH, load_ips
from bairy.device import configs as device_configs
from bairy.device.dash_app import resample_df


def load_data():
  """Load cached data."""
  if os.path.exists(IP_PATH):
    ip_addresses = load_ips()
  else:
    return None

  dfs = glob.glob(HUB_DATA_DIR + '/*.csv')
  names = [f.split('/')[-1].split('.')[0] for f in dfs]

  if 'self' in ip_addresses and os.path.exists(device_configs.DATA_PATH):
    dfs.append(device_configs.DATA_PATH)
    names.append(device_configs.load_device().name)
  if dfs == []:
    return None

  dfs = [pd.read_csv(f) for f in dfs]
  dfs = [df[['time', 'pm_2.5']] for df in dfs]
  dfs = [df.set_index(pd.to_datetime(df['time'])) for df in dfs]
  dfs = [df.drop('time', axis=1) for df in dfs]
  # TODO: give better name
  dfs = [df.rename(columns={'pm_2.5': name}) for df, name in zip(dfs, names)]

  df = pd.concat(dfs)
  # start = pd.Timestamp.now() - pd.Timedelta('1 day')
  # df = df[df.index > start]
  df = resample_df(df)
  return df


def create_fig():
  """Create plotly figure using one or two y-axes."""
  df = load_data()
  if df is None:
    return px.line()
  fig = px.line(df, x=df.index, y=df.columns)
  fig.update_layout(height=800)
  return fig


def serve_plot():
  """Serve dash app plot."""
  fig = create_fig()
  fig.layout.title = 'Data from last 24 hours'
  fig.layout.xaxis.rangeslider.visible = True

  return html.Div(children=[
      html.H1(children='bairy'),
      html.Div(children='Display sensor data from Raspberry Pi.'),
      dcc.Graph(id='graph_day', figure=fig)])


css = 'https://codepen.io/chriddyp/pen/bWLwgP.css'
dash_plot = Dash(requests_pathname_prefix='/plot/', external_stylesheets=[css])
dash_plot.layout = serve_plot



"""Test all device app endpoints."""

import os
from io import BytesIO
import pandas as pd
from fastapi.testclient import TestClient
from bairy.device.app import app
from bairy.device.configs import DATA_PATH


client = TestClient(app)


def test_data_path_exists():
  """Confirm that the device has saved data."""
  assert os.path.exists(DATA_PATH)


def test_all_endpoint_status():
  """Test all endpoints status code through openapi.json json."""
  r = client.get('/openapi.json')
  assert r.status_code == 200
  for e in r.json()['paths'].keys():
    r = client.get(e)
    assert r.status_code == 200

  for e in ['plot', 'table']:
    r = client.get(e)
    assert r.status_code == 200


def test_root():
  """Test root endpoint."""
  r = client.get('/')
  # assert redirecting to plot
  assert 'plot' in r.url.split('/')


def test_data():
  """Test data endpoint."""
  r = client.get('/data')
  df = pd.read_csv(BytesIO(r.content))

  r = client.get('status.json')
  d = r.json()
  headers = d['latest_reading'].keys()
  n_rows = d['data_details']['n_rows']

  assert list(df.columns) == list(headers)
  assert len(df) == n_rows - 1  # don't want to count headers


def test_logs():
  """Test logs endpoint."""
  r = client.get('/logs')
  assert 'INFO' in r.text
  assert 'Uvicorn' in r.text


def test_status():
  """Test status endpoint."""
  r1 = client.get('/status')
  r2 = client.get('/status.json')
  assert r1.json() == r2.json()
  d = r1.json()
  assert 'device_configs' in d
  assert 'latest_reading' in d
  assert 'time' in d['latest_reading']


def test_fake_endpoint():
  """Ensure a fake endpoint returns 404."""
  r = client.get('/fake')
  assert r.status_code == 404



"""FastAPI app to display device data."""

import json
from fastapi import FastAPI
from fastapi.responses import PlainTextResponse, RedirectResponse
from fastapi.middleware.wsgi import WSGIMiddleware
import uvicorn
from bairy.hub.configs import LOG_PATH
from bairy.hub.request import get_all_statuses
from bairy.hub.dash_plot import dash_plot
from bairy.device.app import configure_app_logging


configure_app_logging(LOG_PATH)
app = FastAPI()
app.mount('/plot', WSGIMiddleware(dash_plot.server))


@app.get('/')
async def root():
  """Redirect to plot."""
  return RedirectResponse(url='/plot')


@app.get('/status', response_class=PlainTextResponse)
async def status():
  """Get status of each device."""
  statuses = get_all_statuses()
  return json.dumps(statuses, indent=4)


@app.get('/logs', response_class=PlainTextResponse)
async def logs():
  """Return app log as plain text."""
  with open(LOG_PATH) as f:
    return f.read()


def run_app():
  """Run app as separate process."""
  uvicorn.run(app, host='0.0.0.0', port=8000)



"""Install bairy."""
import setuptools

with open('README.md') as f:
  long_description = f.read()
description = 'Display data from Raspberry Pi.'


install_requires = [
    'uvicorn',
    'fastapi',
    'pydantic',
    'dash',
    'plotly',
    'pandas',
    'nest_asyncio',
    'aiohttp',
    'smbus2',
    'gpiozero'
]

setuptools.setup(
    name='bairy',
    author='Zeb Engberg',
    author_email='zebengberg@gmail.com',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/zebengberg/bairy',
    packages=setuptools.find_namespace_packages(exclude=['tests*']),
    python_requires='>=3.7.0',
    install_requires=install_requires,
    entry_points={'console_scripts': ['bairy = bairy.__main__:main']},
    version='0.0.6',
    classifiers=['License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.7',
                 'Topic :: Internet :: WWW/HTTP :: WSGI',
                 'Topic :: Scientific/Engineering :: Visualization'],
    license='MIT')



"""A module holding utility functions for device."""

from __future__ import annotations
import logging
import os
import socket
from bairy.device.configs import DATA_PATH, LOG_FORMAT, DATE_FORMAT


def configure_logging(log_path: str):
  """Set up root logger to send logs to file and console."""
  logger = logging.getLogger()
  logger.setLevel(logging.INFO)
  formatter = logging.Formatter(fmt=LOG_FORMAT, datefmt=DATE_FORMAT)
  file_handler = logging.FileHandler(log_path)
  file_handler.setFormatter(formatter)
  logger.addHandler(file_handler)
  console_handler = logging.StreamHandler()
  console_handler.setFormatter(formatter)
  logger.addHandler(console_handler)


def read_headers():
  """Read first line of data."""
  with open(DATA_PATH) as f:
    headers = f.readline()
    return headers.rstrip()


def read_last_line():
  """Read last line of data."""
  # fast approach to get final line of a file
  # see https://stackoverflow.com/questions/46258499/
  with open(DATA_PATH, 'rb') as f:
    f.seek(-2, os.SEEK_END)
    while f.read(1) != b'\n':
      f.seek(-2, os.SEEK_CUR)
    return f.readline().decode()


def latest_data():
  """Get last line of data as dictionary."""
  last_line = read_last_line()
  values = last_line.split(',')
  time = values.pop(0)
  values = [int(v) for v in values]

  d: dict[str, str | int] = {'time': time}
  headers = read_headers().split(',')[1:]
  d.update(dict(zip(headers, values)))
  return d


def get_data_size():
  """Return the size of the data file as a string."""
  n = os.path.getsize(DATA_PATH)
  for unit in ['', 'Ki', 'Mi', 'Gi']:
    if n < 1024.0:
      return f'{n:.2f} {unit}B'
    n /= 1024.0
  raise OverflowError


def count_rows(path: str):
  """Count number of rows in CSV at path."""
  assert path[-4:] == '.csv'
  with open(path) as f:
    return sum(1 for _ in f)


def print_local_ip_address():
  """See https://stackoverflow.com/questions/166506/"""

  s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
  s.connect(('8.8.8.8', 80))
  ip = s.getsockname()[0]
  s.close()
  print('#' * 65)
  print('LOCAL IP ADDRESS:', ip)
  print('#' * 65)
